import pathlib
from collections.abc import Iterable, Sequence
from enum import Enum, auto

import cv2
import mss
import mss.screenshot
import numpy


class SortBy(Enum):
    CONFIDENCE = auto()
    Y = auto()
    X = auto()


def _get_base_image(
    sub_image: str | pathlib.Path | numpy.ndarray,
    image: str | pathlib.Path | numpy.ndarray
) -> numpy.ndarray:
    sub_image = to_ndarray(sub_image)
    image = to_ndarray(image)

    width = max(sub_image.shape[1], image.shape[1])
    height = max(sub_image.shape[0], image.shape[0])
    image_ = numpy.zeros((height, width, 3), dtype=numpy.uint8)
    image_[0:image.shape[0], 0:image.shape[1]] = image

    return image_


def _swap_axis_in_place(array: numpy.ndarray):
    array[:, 0], array[:, 1] = array[:, 1], array[:, 0].copy()


def compare(
    a: str | pathlib.Path | numpy.ndarray,
    b: str | pathlib.Path | numpy.ndarray,
    mask: str | pathlib.Path | numpy.ndarray = None
) -> float:
    _, score, _, _ = cv2.minMaxLoc(match(a, b, mask))
    return score


def get_all_positions_in_image(
    sub_images: str | pathlib.Path | numpy.ndarray | Iterable[str | pathlib.Path | numpy.ndarray],
    image: str | pathlib.Path | numpy.ndarray,
    mask: str | pathlib.Path | numpy.ndarray = None,
    confidence=1.0,
    center=False,
    sort_by: SortBy | Iterable[SortBy] = None
) -> numpy.ndarray:
    def add_confidence(array: numpy.ndarray) -> numpy.ndarray:
        return numpy.array([*array, 1 - matches[*array]])

    def center_points(array: numpy.ndarray) -> numpy.ndarray:
        # noinspection PyTypeChecker
        return numpy.array([*get_center(array[0], array[1], sub_image.shape[0], sub_image.shape[1]), array[2]])

    def process_sort_by(sort_by_):
        match sort_by_:
            case None:
                new_sort_by = list(SortBy)
            case [*_]:
                new_sort_by = []
                for category in sort_by_:
                    if category not in new_sort_by:
                        new_sort_by.append(category)
            case _:
                new_sort_by = [sort_by_]
        for category in iter(SortBy):
            if category not in new_sort_by:
                new_sort_by.append(category)
        return new_sort_by

    if isinstance(sub_images, str | pathlib.Path | numpy.ndarray):
        sub_images = (sub_images,)
    sort_by = process_sort_by(sort_by)

    all_points = numpy.empty((0, 3), dtype=float)
    for sub_image in sub_images:
        sub_image = to_ndarray(sub_image)
        matches = match(sub_image, image, mask)
        points = numpy.argwhere(matches >= confidence)
        if not points.size:
            continue
        points = numpy.apply_along_axis(add_confidence, 1, points)
        if center:
            points = numpy.apply_along_axis(center_points, 1, points)
        all_points = numpy.concatenate((all_points, points))

    _swap_axis_in_place(all_points)
    all_points = numpy.core.records.fromarrays(all_points.transpose(), names='x, y, confidence', formats='int, int, float')

    if sort_by != [SortBy.Y, SortBy.X, SortBy.CONFIDENCE]:
        all_points.sort(order=[category.name.lower() for category in sort_by])

    # restore confidence value for reverse sort
    if all_points.size:
        all_points['confidence'] = 1 - all_points['confidence']

    return all_points


def get_all_positions_in_screen(
    sub_images: str | pathlib.Path | numpy.ndarray | Iterable[str | pathlib.Path | numpy.ndarray],
    mask: str | pathlib.Path | numpy.ndarray = None,
    region: Sequence[int, int, int, int] = None,
    confidence=1.0,
    center=False,
    region_relative_coordinates=False,
    sort_by: SortBy | Iterable[SortBy] = None
) -> tuple[numpy.ndarray, numpy.ndarray]:
    image, (left, top) = get_screenshot(region)
    points = get_all_positions_in_image(sub_images, image, mask, confidence, center, sort_by)

    if points.size and not region_relative_coordinates:
        points['x'] += left
        points['y'] += top

    return points, image


def get_center(x: int, y: int, width: int, height: int) -> tuple[int, int]:
    return x + round(width / 2), y + round(height / 2)


def get_screenshot(
    region: Sequence[int, int, int, int] = None,
    capturer: mss.base.MSSBase = None
) -> tuple[numpy.ndarray, tuple[int, int]]:
    def get_mss_region(region_: Sequence[int, int, int, int] = None) -> dict[str, int]:
        if region_:
            return {
                "left": region_[0],
                "top": region_[1],
                "width": region_[2] - region_[0],
                "height": region_[3] - region_[1],
            }
        else:
            return capturer.monitors[0]

    if capturer:
        region = get_mss_region(region)
        screenshot = capturer.grab(region)
    else:
        with mss.mss() as capturer:
            region = get_mss_region(region)
            screenshot = capturer.grab(region)

    return to_ndarray(screenshot), (region['left'], region['top'])


def is_region_color(
    region: Sequence[int, int, int, int],
    target_color: tuple[int, int, int],
    tolerance: int
) -> bool:
    color = region_color_mean(region)
    return (
        abs(target_color[0] - color[0]) < tolerance
        and
        abs(target_color[1] - color[1]) < tolerance
        and
        abs(target_color[2] - color[2]) < tolerance
    )


def match(
    sub_image: str | pathlib.Path | numpy.ndarray,
    image: str | pathlib.Path | numpy.ndarray,
    mask: str | pathlib.Path | numpy.ndarray = None,
) -> numpy.ndarray:
    sub_image = to_ndarray(sub_image)
    image = _get_base_image(sub_image, image)

    kwargs = {}
    if mask:
        kwargs['mask'] = to_ndarray(mask)

    return cv2.matchTemplate(image, sub_image, cv2.TM_CCOEFF_NORMED, **kwargs)


def region_color_mean(region: Sequence[int, int, int, int]) -> tuple[int, int, int]:
    image, _ = get_screenshot(region)
    image = image.astype(numpy.uint16)

    red_sum = (image ** 2)[:, :, 0].sum(dtype=numpy.uint64)
    green_sum = (image ** 2)[:, :, 1].sum(dtype=numpy.uint64)
    blue_sum = (image ** 2)[:, :, 2].sum(dtype=numpy.uint64)

    n_pixels = image.shape[0] * image.shape[1]

    return (
        round((red_sum / n_pixels) ** (1 / 2)),
        round((green_sum / n_pixels) ** (1 / 2)),
        round((blue_sum / n_pixels) ** (1 / 2))
    )


def search_in_image(
    sub_image: str | pathlib.Path | numpy.ndarray,
    image: str | pathlib.Path | numpy.ndarray,
    mask: str | pathlib.Path | numpy.ndarray = None,
    confidence=1.0,
    center=False
) -> tuple[int, int] | None:
    sub_image = to_ndarray(sub_image)
    matches = match(sub_image, image, mask)
    _, confidence_, _, (x, y) = cv2.minMaxLoc(matches)
    if confidence_ < confidence:
        return

    if center:
        x, y = get_center(x, y, sub_image.shape[1], sub_image.shape[0])

    return x, y


def search_in_screen(
    sub_image: str | pathlib.Path | numpy.ndarray,
    mask: str | pathlib.Path | numpy.ndarray = None,
    region: Sequence[int, int, int, int] = None,
    confidence=1.0,
    center=False,
    region_relative_coordinates=False
) -> tuple[tuple[int, int], numpy.ndarray] | None:
    image, (left, top) = get_screenshot(region)
    if not (point := search_in_image(
        sub_image,
        image,
        mask,
        confidence,
        center
    )):
        return

    if not region_relative_coordinates:
        point = (point[0] + left, point[1] + top)

    return point, image


def to_ndarray(image: str | pathlib.Path | mss.screenshot.ScreenShot | numpy.ndarray) -> numpy.ndarray:
    match image:
        case str() | pathlib.Path():
            image = cv2.imread(str(image))
        case mss.screenshot.ScreenShot():
            image = numpy.array(image)
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    return image[:, :, :3]



import datetime
import json
import numbers as numbers_module
import pathlib
import random
import re
import secrets
import string
from typing import Iterable, Type, overload

import jellyfish
import unicodedata

from flanautils import constants, iterables


def cartesian_product_string_matching(a_text: str | Iterable[str], b_text: str | Iterable[str], min_score: float = 0) -> dict[str, dict[str, float]]:
    """
    Compare between all the strings of the first iterable with all of the second (cartesian product) and returns a
    dictionary with the scores.
    """

    a_words = a_text.split() if isinstance(a_text, str) else a_text
    b_words = b_text.split() if isinstance(b_text, str) else b_text
    return {a_word: matches for a_word in a_words if (matches := {b_word: score for b_word in b_words if (score := jellyfish.jaro_winkler_similarity(a_word, b_word)) >= min_score})}


@overload
def cast_number(x: numbers_module.Number, raise_exception=True) -> numbers_module.Number:
    pass


@overload
def cast_number(x: str, raise_exception=True) -> numbers_module.Number | str:
    pass


@overload
def cast_number(x: bytes, raise_exception=True) -> numbers_module.Number | bytes:
    pass


def cast_number(x: numbers_module.Number | str | bytes, raise_exception=True) -> numbers_module.Number | str | bytes:
    """
    Try to cast x to a number.

    If raise_exception=False (True by default) and the casting fails, returns the input as it is.
    """

    if isinstance(x, numbers_module.Number):
        return x

    try:
        return int(x)
    except ValueError:
        try:
            return float(x)
        except ValueError:
            if raise_exception:
                raise
            else:
                return x


def find_coordinates(text: str) -> tuple[float, float] | list[tuple[float, float]] | None:
    """
    Find coordinates in string.

    - If one pair of numbers are matched, returns (latitude, longitude)
    - If more of one pair of numbers are matched, returns [(latitude, longitude), ...]
    - If not match, returns None

    >>> find_coordinates('holaaaa 85.1, 44.3515')
    (85.1, 44.3515)
    >>> find_coordinates('ey que pasa+654.1464;,;+ + +-456.4164hola')
    (654.1464, -456.4164)
    >>> find_coordinates('ey que pasa+654.1464;,;+ + -+456.4164hola    40+5.5')
    [(654.1464, 456.4164), (40.0, 5.5)]
    >>> find_coordinates('ey que pasa')

    """

    results = re.findall(r'[-+\d.]+[,;\s+-]+[-+\d.]+', replace(text, {'-': ' -', '+': ' +'}))

    formatted_results = []
    for result in results:
        try:
            words = re.split(r'[,;\s+]+', result)
            latitude, longitude = iterables.filter(words, int | float, cast_numbers=True)
        except ValueError:
            continue
        formatted_results.append((float(latitude), float(longitude)))

    match formatted_results:
        case [single]:
            return single
        case [_, *_]:
            return formatted_results


def find_jsons(text: str | pathlib.Path) -> list[dict]:
    """Find all well formatted JSONs in a string or in a pathlib.Path and return them in dictionaries."""

    text = read_file(text) or text

    jsons = []
    position = 0
    while (position := text.find('{', position)) != -1:
        result, position = json.JSONDecoder().raw_decode(text, position)
        jsons.append(result)
    return jsons


def find_environment_variables(text: str | pathlib.Path) -> dict:
    """Looks for environment variables in a string or in a pathlib.Path in .env format (key=value)."""

    text = read_file(text) or text
    # noinspection PyTypeChecker
    return dict(line.split('=', maxsplit=1) for original_line in text.splitlines() if '=' in (line := original_line.strip()) and not line.startswith('#'))


def find_urls(text: str) -> list[str]:
    """Returns the substrings that match the url pattern."""

    return re.findall(r'(?:http|www\.)[-a-zA-Z0-9()@:%_+.~#?&/=]+', text)


def find_url_domains(text: str) -> list[str]:
    """Returns the substrings that match the url domain pattern."""

    return re.findall(r'(?:http.+?/|www\.|@)([\w.]+)(?!.*@)', text)


def join_last_separator(elements: Iterable, separator: str, last_separator: str, final_char='') -> str:
    """
    Join all the elements in a string, using a separator for all of them except the last one, where it uses
    last_separator.

    You can also add to final char.

    >>> join_last_separator(['Uno', 'dos', 'tres', 'cuatro'], ', ', ' y ', '.')
    'Uno, dos, tres y cuatro.'
    """

    elements = list(elements)
    if not len(elements):
        return ''
    if len(elements) == 1:
        return f'{elements[0]}{final_char}'
    return f'{separator.join(elements[:-1])}{last_separator}{elements[-1]}{final_char}'


def numbers_to_text(number: int, language='es') -> str:
    """
    Convert an integer into its textual representation.

    >>> numbers_to_text(7)
    'siete'
    >>> numbers_to_text(92)
    'noventa y dos'
    >>> numbers_to_text(15)
    'quince'
    >>> numbers_to_text(16)
    'dieciséis'
    """

    if number >= 0:
        is_positive = True
    else:
        is_positive = False
        number *= -1

    if number not in range(0, 101):
        raise ValueError('value out of range')

    if language == 'es':
        number_word_es = constants.NUMBER_WORDS[language]
        try:
            result = number_word_es[number]
        except KeyError:
            tens, units = divmod(number, 10)
            separator = ''
            if tens == 1:
                tens_word = 'dieci'
            elif tens == 2:
                tens_word = 'veinti'
            elif tens == 10:
                tens_word = 'ciento'
            else:
                separator = ' y '
                tens_word = number_word_es[number - units]
            result = f"{tens_word}{separator}{number_word_es[units]}"
        if is_positive:
            return result
        else:
            return f"{number_word_es['-']} {result}"
    else:
        raise NotImplementedError('not implemented for that language')


@overload
def random_string(min_len: int, max_len: int, letters=True, numbers=True, n_spaces=0) -> str:
    pass


@overload
def random_string(str_len=10, letters=True, numbers=True, n_spaces=0) -> str:
    pass


def random_string(min_len=10, max_len=None, letters=True, numbers=True, n_spaces=0) -> str:
    """
    Generate a random string.

    It can generate random strings of random size within the limits specified by arguments.

    You can specify the content of the strings: letters, numbers and number of spaces.
    """

    if isinstance(min_len, bool):
        letters = min_len
        min_len = None
        if isinstance(max_len, bool):
            numbers = max_len
            max_len = None
    elif isinstance(max_len, bool):
        letters = max_len
        max_len = None
    characters = f"{string.ascii_letters if letters else ''}{string.digits if numbers else ''}{' ' * n_spaces}"

    if max_len is not None:
        if max_len < min_len:
            return ''
        min_len = random.randint(min_len, max_len)

    return ''.join(secrets.choice(characters) for _ in range(min_len))


def read_file(path: str | pathlib.Path) -> str | None:
    """
    Returns the content of the file as text.

    Returns None if path is not a valid path string or pathlib.Path.
    """

    match path:
        case str() | pathlib.Path() as path:
            if pathlib.Path(path).is_file():
                with open(path) as file:
                    return file.read()


def remove_accents(text: str, ignore: Iterable = ('ñ', 'ç')) -> str:
    """
    Removes the accents from the string.

    Ignore the characters contained in ignore argument. By default ignore=('ñ', 'ç').

    >>> remove_accents('aáeèiîoöuuºªçñ')
    'aaeeiioouuºªçñ'
    >>> remove_accents('👉🏻Mañana iba a salir pero el otro día iba por la calle y casi me atropella un camión 🚛 que iba muy rápido.')
    '👉🏻Mañana iba a salir pero el otro dia iba por la calle y casi me atropella un camion 🚛 que iba muy rapido.'
    """

    def remove_accents_generator():
        for char in text:
            if char in ignore:
                yield char
            else:
                normalize_chars = unicodedata.normalize('NFD', char)
                for normalize_char in normalize_chars:
                    if unicodedata.category(normalize_char) != 'Mn':
                        yield normalize_char

    return ''.join(remove_accents_generator())


def remove_symbols(text: str, ignore: Iterable = (), replace_with='') -> str:
    """
    Remove the typical symbols from the string.

    Ignore the characters contained in ignore argument. By default ignore=().

    Symbols are replaced with replace_with argument. By default replace_with=''.

    >>> remove_symbols('¿No vás muy rápido? Yo creo que sí.')
    'No vás muy rápido Yo creo que sí'
    >>> remove_symbols('¿No vás muy rápido? Yo creo que sí.', ignore=('?',), replace_with='8')
    '8No vás muy rápido? Yo creo que sí8'
    """

    return ''.join((char if char in ignore else replace_with) if char in constants.SYMBOLS else char for char in text)


replace_symbols = remove_symbols


def replace(text: str, replacements: dict) -> str:
    """
    Returns a copy of text with the replacements applied.

    >>> replace('abc.-.', {'.': '*', '-': '<===>'})
    'abc*<===>*'
    >>> replace('Hola1 que2 ase3', {'Hola': 'Hi', 'que': 'ase', 'ase': None})
    'Hi1 ase2 3'
    """

    if not replacements:
        return text

    result: list[str] = []

    start = 0
    while start < len(text):
        pattern: list[str] = []

        for char in text[start:]:
            pattern.append(char)
            try:
                new_value = replacements[''.join(pattern)]
            except KeyError:
                pass
            else:
                result.append(new_value or '')
                start += len(pattern)
                break
        else:
            result.append(text[start])
            start += 1

    return ''.join(result)


translate = replace


@overload
def str_to_class(class_names: str | Type, globals_: dict) -> Type:
    pass


@overload
def str_to_class(class_names: Iterable[str | Type], globals_: dict) -> tuple[Type, ...]:
    pass


def str_to_class(class_names: str | Type | Iterable[str | Type], globals_: dict) -> Type | tuple[Type, ...]:
    """Converts the class or classes names into Python Types found in globals_."""

    globals_ = globals_ or {}
    if isinstance(class_names, str):
        return globals_[class_names]
    else:
        return tuple(globals_[class_name] for class_name in class_names)


def text_to_number(text: str, ignore_no_numbers=True, language='es') -> int | float:
    """
    Convert numbers from text to numeric and return the sum of them.

    >>> text_to_number('cientocincuentaydos')
    152
    >>> text_to_number('Borra veintidos mensajes. Y... luego borra otros treinta y cinco.')
    57
    >>> text_to_number('uno menos dos menos tres')
    -4
    >>> text_to_number('uno - dos y tres')
    -4
    >>> text_to_number('uno menos dos mas tres')
    2
    >>> text_to_number('uno - dos mas tres')
    2
    >>> text_to_number('uno - dos + tres')
    2
    >>> text_to_number('Uno más dos. Y luego cuarenta y cuatro y cuarenta y 2 y 10 y 1 más 7.')
    107
    >>> text_to_number('0.1 + uno más dos, 5.2 -1.1 mas 5.3 -5.8 ')
    6.7
    """

    text = remove_accents(text)
    text = remove_symbols(text, ('+', '-', '.'))

    if language == 'es':
        number_words_es = constants.NUMBER_WORDS[language]
        text = replace(text, {'y': ' ', 'ç': None})
        text = re.sub(r'(([ei]nt[aeio])|(ec))', r'\1 ', text)
        words = text.lower().split()
        n = 0
        sign = 1
        for word in words:
            word = word.strip('.')
            try:
                n += sign * cast_number(word)
            except ValueError:
                pass
            else:
                continue

            if len(word) > constants.TEXT_TO_NUMBER_MAX_WORD_LENGTH:
                continue

            if word == '+' or jellyfish.jaro_winkler_similarity(word, number_words_es['+']) >= constants.NUMBERS_SCORE_MATCHING:
                sign = 1
                continue
            elif word == '-' or jellyfish.jaro_winkler_similarity(word, number_words_es['-']) >= constants.NUMBERS_SCORE_MATCHING:
                sign = -1
                continue

            if word_matches := cartesian_product_string_matching(word, number_words_es.values(), constants.NUMBERS_SCORE_MATCHING):
                number_word = max(word_matches[word].items(), key=lambda item: item[1])[0]
                n += sign * number_words_es[number_word]
            elif not ignore_no_numbers:
                raise KeyError(word)

        return n

    raise NotImplementedError('not implemented for that language')


def text_to_time(text: str | Iterable[str], language='es') -> datetime.timedelta:
    """
    Convert time in textual representation into a datetime.timedelta.

    >>> text_to_time('Un minuto y 10 segundos.')
    datetime.timedelta(seconds=70)
    """

    if isinstance(text, str):
        words = text.split()
    else:
        words = text

    delta_time = datetime.timedelta()
    n = 0

    if language == 'es':
        for word in words:
            if jellyfish.jaro_winkler_similarity(word, 'segundo') >= constants.TIME_UNITS_SCORE_MATCHING:
                delta_time += datetime.timedelta(seconds=n)
                n = 0
            elif jellyfish.jaro_winkler_similarity(word, 'minuto') >= constants.TIME_UNITS_SCORE_MATCHING or jellyfish.jaro_winkler_similarity(word, 'min') >= constants.TIME_UNITS_SCORE_MATCHING:
                delta_time += datetime.timedelta(minutes=n)
                n = 0
            elif jellyfish.jaro_winkler_similarity(word, 'hora') >= constants.TIME_UNITS_SCORE_MATCHING:
                delta_time += datetime.timedelta(hours=n)
                n = 0
            elif jellyfish.jaro_winkler_similarity(word, 'dia') >= constants.TIME_UNITS_SCORE_MATCHING:
                delta_time += datetime.timedelta(days=n)
                n = 0
            elif jellyfish.jaro_winkler_similarity(word, 'semana') >= constants.TIME_UNITS_SCORE_MATCHING:
                delta_time += datetime.timedelta(weeks=n)
                n = 0
            elif jellyfish.jaro_winkler_similarity(word, 'mes') >= constants.TIME_UNITS_SCORE_MATCHING:
                delta_time += datetime.timedelta(weeks=n * constants.WEEKS_IN_A_MONTH)
                n = 0
            elif jellyfish.jaro_winkler_similarity(word, 'año') >= constants.TIME_UNITS_SCORE_MATCHING:
                delta_time += datetime.timedelta(weeks=n * constants.WEEKS_IN_A_YEAR)
                n = 0
            else:
                n += text_to_number(word)

        return delta_time

    raise NotImplementedError('not implemented for that language')



import asyncio
import functools
import inspect
import timeit
from typing import Any, Callable, Iterable, Type

from flanautils import iterables


def is_function(func: Any) -> bool:
    """Checks if the func object is considered a function."""

    if isinstance(func, functools.partial):
        func = func.func

    return inspect.isfunction(func) or inspect.ismethod(func)


# --------------------------------------------------------- #
# -------------------- META DECORATORS -------------------- #
# --------------------------------------------------------- #
def shift_args_if_called(func_: Callable = None, *, exclude_self_types: str | Type | Iterable[str | Type] = (), globals_: dict = None) -> Callable:
    """Decorator for decorators that shifts the arguments depending on whether the decorator is called or not."""

    if func_ is not None and not is_function(func_):
        func_, exclude_self_types, globals_ = iterables.shift_function_args(func_, exclude_self_types, globals_, func=shift_args_if_called)

    def decorator(func: Callable):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            self, args = iterables.separate_self_from_args(args, exclude_self_types, globals_)
            if args and args[0] is not None and not is_function(args[0]):
                args = iterables.shift_function_args(*args, func=func)

            if self:
                return func(self, *args, **kwargs)
            else:
                return func(*args, **kwargs)

        return wrapper

    return decorator(func_) if func_ else decorator


# ---------------------------------------------------- #
# -------------------- DECORATORS -------------------- #
# ---------------------------------------------------- #
@shift_args_if_called
def repeat(func_: Callable = None, /, times=2) -> Callable:
    """
    Decorator that makes the decorated function be executed the specified number of times (by default times=2).

    Returns the last result.
    """

    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> Any:
            result = None
            for _ in range(int(times)):
                result = func(*args, **kwargs)
            return result

        return wrapper

    return decorator(func_) if func_ else decorator


@shift_args_if_called
def return_if_first_empty(func_: Callable = None, /, return_: Any = None, exclude_self_types: str | Type | Iterable[str | Type] = (), globals_: dict = None) -> Callable:
    """
    Decorator that aborts the execution of the function if the boolean value of the first element is False.

    In case of cancellation returns the value provided by return_ (by default return_=None).

    Ignore the first element if it is an instance of exclude_self_types.
    """

    return_ = return_() if callable(return_) else return_

    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> Any:
            self, args = iterables.separate_self_from_args(args, exclude_self_types, globals_)

            if not args[0] if args else not next(iter(kwargs.values()), None):
                if asyncio.iscoroutinefunction(func):
                    async def temp():
                        return return_

                    return temp()
                return return_

            if self:
                args = (self, *args)
            return func(*args, **kwargs)

        return wrapper

    return decorator(func_) if func_ else decorator


@shift_args_if_called
def time_it(func_: Callable = None, /, n_executions=1) -> Callable:
    """Decorator that prints the seconds it takes for the function to run."""

    def decorator(func: Callable):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            new_template = f"""def inner(_it, _timer{{init}}):
                                   {{setup}}
                                   _t0 = _timer()
                                   for _i in _it:
                                       result = {{stmt}}
                                   _t1 = _timer()
                                   print({repr(func.__name__)} + ': ' + str(_t1 - _t0) + ' seconds')
                                   return result"""

            timeit.template = new_template
            return timeit.timeit(lambda: func(*args, **kwargs), number=n_executions)

        return wrapper

    return decorator(func_) if func_ else decorator



from dataclasses import dataclass
from typing import Generic, TypeVar

from flanautils.models.bases import FlanaBase

T = TypeVar('T')


@dataclass(frozen=True)
class ScoreMatch(FlanaBase, Generic[T]):
    """Dataclass to store the relationship between an element and a score."""

    element: T
    score: float

    def __add__(self, other):
        other_value = other.score if isinstance(other, ScoreMatch) else other
        return self.score + other_value

    def __radd__(self, other):
        return self + other

    def __eq__(self, other):
        if isinstance(other, ScoreMatch):
            return (self.element, self.score) == (other.element, other.score)
        else:
            return self.element == other

    def __lt__(self, other):
        return (other.score, str(self.element)) < (self.score, str(other.element))



import random
from typing import Iterable

from flanautils import strings


def random_elements(k=1) -> list:
    return [random.choice(
        (
            random.randint(-150, 150),
            random.uniform(-150, 150),
            strings.random_string(0, 5)
        )
    ) for _ in range(k)]


def random_collections(k=1) -> list:
    return [random.choice(
        (
            *random_elements(random.randint(0, 3)),
            (*random_elements(random.randint(0, 3)),),
            [*random_elements(random.randint(0, 3))],
            {*random_elements(random.randint(0, 3))}
        )
    ) for _ in range(k)]


def list_without_repetitions(*iterables: Iterable[Iterable]) -> list:
    result = []
    for iterable in iterables:
        for element in iterable:
            if element not in result:
                result.append(element)

    return result



import inspect
import typing
from typing import Any, Callable, Iterable, Iterator, Sequence, Type, overload

from flanautils import maths, strings


def chunks(elements: Sequence, size, lazy=False) -> Iterator[Sequence] | list[Sequence]:
    """Create successive n-sized chunks from elements."""

    generator_ = (elements[i:i + size] for i in range(0, len(elements), size))
    if lazy:
        return generator_
    else:
        return list(generator_)


# noinspection PyShadowingNames, PyShadowingBuiltins
def filter(
    elements: Iterable,
    target: Any = None,
    condition: Callable[..., bool] = None,
    cast_numbers=False,
    lazy=False
) -> Iterator | list:
    """
    Smart function that find anything in an iterable (classes, objects, ...).

    If condition is not None, return the all elements that matches it.

    >>> elements = [1, 2.2, '3', 4, 'hola', '6.6']

    >>> filter(elements, 1)
    [1]
    >>> filter(elements, int)
    [1, 4]
    >>> filter(elements, float)
    [2.2]
    >>> filter(elements, 6.6, cast_numbers=True)
    [6.6]
    >>> filter(elements, int | float, cast_numbers=True)
    [1, 2.2, 3, 4, 6.6]
    >>> import numbers  # Python Standard Library
    >>> filter(elements, numbers.Number, cast_numbers=True)
    [1, 2.2, 3, 4, 6.6]
    >>> filter(elements, condition=lambda e: isinstance(e, int | float) and e > 3, cast_numbers=True)
    [4, 6.6]
    >>> type(filter(elements, int | float, cast_numbers=True, lazy=True))
    <class 'generator'>
    >>> list(filter(elements, int | float, cast_numbers=True, lazy=True))
    [1, 2.2, 3, 4, 6.6]
    """

    if condition is None:
        if isinstance(target, type) or isinstance(typing.get_origin(target), type | type(typing.Union)):
            def condition(element: Any) -> bool:
                try:
                    return issubclass(element, target)
                except TypeError:
                    return isinstance(element, target)
        else:
            def condition(element: Any) -> bool:
                return element == target

    if cast_numbers:
        generator_ = (final_element for element in elements if condition(final_element := strings.cast_number(element, raise_exception=False)))
    else:
        generator_ = (element for element in elements if condition(element))

    return generator_ if lazy else list(generator_)


def filter_exceptions(elements: Iterable) -> tuple[list, list[Exception]]:
    """Filters the exceptions of the iterable and returns a tuple with the separated results."""

    exceptions = []
    no_exceptions = []

    for element in elements:
        if isinstance(element, Exception):
            exceptions.append(element)
        else:
            no_exceptions.append(element)

    return no_exceptions, exceptions


# noinspection PyShadowingNames
def find(elements: Iterable, target: Any = None, condition: Callable[..., bool] = None, cast_numbers=False) -> Any:
    """
    Smart function that find anything in an iterable (classes, objects, ...).

    If condition is not None, return the first element that matches it.

    Returns None if nothing matches.

    >>> elements = [1, 2, '3', 4, 'hola', '6.6']

    >>> find(elements, 2)
    2
    >>> find(elements, int)
    1
    >>> find(elements, float)

    >>> find(elements, 6.6, cast_numbers=True)
    6.6
    >>> find(elements, float, cast_numbers=True)
    6.6
    """

    return next(filter(elements, target, condition, cast_numbers, lazy=True), None)


def flatten(*args: Iterable, depth: int = None, lazy=False) -> Iterator | list:
    """
    Iterates and flattens iterables recursively according to the specified depth.

    If depth=None (the default) it flattens recursively until it finds no iterable.

    >>> type(flatten([1, 2, [3, 4, ['cinco']]]))
    <class 'list'>
    >>> type(flatten([1, 2, [3, 4, ['cinco']]], lazy=True))
    <class 'generator'>
    >>> flatten([1, 2, [3, 4, ['cinco']]])
    [1, 2, 3, 4, 'cinco']
    >>> flatten([1, 2, [3, 4, ['cinco']]], depth=1)
    [1, 2, [3, 4, ['cinco']]]
    >>> flatten([1, 2, [3, 4, ['cinco']]], depth=2)
    [1, 2, 3, 4, ['cinco']]
    """

    current_depth = -1

    def flatten_generator(*args_: Iterable, depth_=None) -> Iterator:
        nonlocal current_depth

        if depth_ is not None:
            current_depth += 1

        for arg_ in args_:
            if (
                isinstance(arg_, Iterable)
                and
                not isinstance(arg_, (str, bytes))
                and
                (depth_ is None or current_depth < depth_)
            ):
                yield from flatten_generator(*arg_, depth_=depth_)
                if depth_ is not None:
                    current_depth -= 1
            else:
                yield arg_

    generator_ = flatten_generator(*args, depth_=depth)
    return generator_ if lazy else list(generator_)


@overload
def frange(start: float = 0, stop: float = float('inf'), step: float = 1, include_last=False) -> Iterator[float]:
    pass


@overload
def frange(stop: float = 0, include_last=False) -> Iterator[float]:
    pass


def frange(start: float = None, stop: float = None, step: float = 1, include_last=False) -> Iterator[float]:
    """
    Returns a Generator that works like range but with floats.

    Can generate infinite numerical series.

    >>> type(frange(5))
    <class 'generator'>
    >>> list(frange(5))
    [0.0, 1.0, 2.0, 3.0, 4.0]
    >>> list(frange(2, 5))
    [2.0, 3.0, 4.0]
    >>> list(frange(2, 5, 0.5))
    [2.0, 2.5, 3.0, 3.5, 4.0, 4.5]
    >>> list(frange(2, 5, 0.5, include_last=True))
    [2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0]
    >>> list(frange(3, 5, 0.2))
    [3.0, 3.2, 3.4000000000000004, 3.6000000000000005, 3.8000000000000007, 4.000000000000001, 4.200000000000001, 4.400000000000001, 4.600000000000001, 4.800000000000002]
    """

    if step == 0:
        raise ValueError('step must not be zero')

    match start, stop:
        case [(int() | float()) as stop, None]:
            start = 0
        case [(int() | float()) as stop, bool(include_last)]:
            start = 0

    step_sign = maths.sign(step)
    start = 0 if start is None else start
    stop = step_sign * (float('inf') if stop is None else stop)

    i = start
    while step_sign * i < stop or (step_sign * i == stop and include_last):
        yield float(i)
        i += step


def iterate_n(iterable: Iterable, n: int, default: Any = ...) -> Any | None:
    """
    Iterates an iterable a number of times and returns the obtained value.

    If default is given and the iterator is exhausted, it is returned instead of raising StopIteration.
    """

    if n <= 0:
        return

    iterable = iter(iterable)
    if default is not ...:
        def iterate():
            return next(iterable, default)
    else:
        def iterate():
            return next(iterable)

    for _ in range(n - 1):
        iterate()

    return iterate()


def separate_self_from_args(
    args: tuple,
    exclude_self_types: str | Type | Iterable[str | Type],
    globals_: dict = None
) -> tuple[Any, tuple]:
    """
    Given a function arguments iterable: if the first arg is an instance of exclude_self_types return (args[0],
    args[1:]), else (None, args[1:]).

    If exclude_self_types is a string or iterable of strings, strings.str_to_class(exclude_self_types, globals_)
    is called.
    """

    match exclude_self_types:
        case str() | [*_, str()]:
            exclude_self_types = strings.str_to_class(exclude_self_types, globals_)

    if args and isinstance(args[0], exclude_self_types):
        self, *args = args
    else:
        self = None

    return self, args


def shift_function_args(*args, func: Callable = None, first_arg_default: Any = None) -> tuple:
    """
    Shift received arguments one position to the right, taking into account the possible default value of the first
    argument.
    """

    if isinstance(func, Callable):
        first_arg_default = next(iter(inspect.signature(func).parameters.values())).default

    return None if first_arg_default is inspect.Parameter.empty else first_arg_default, *args



from dataclasses import dataclass

from flanautils import constants, strings
from flanautils.models.bases import FlanaBase


def _round_if_close_to_unit(number: float, n_digits: int = 10) -> float:
    rounded = round(number, n_digits)
    integer_part = int(number)
    return float(rounded) if rounded in (integer_part, integer_part + 1) else number


@dataclass(unsafe_hash=True)
class TimeUnits(FlanaBase):
    """
    Represents the time information grouping it into typical units.

    You can represent that tense in textual form according to the language.

    >>> TimeUnits(hours=1000)
    TimeUnits(years=0, months=1, weeks=1, days=4, hours=5, minutes=59, seconds=58.847999999838976)
    >>> TimeUnits(hours=7*24)
    TimeUnits(years=0, months=0, weeks=1, days=0, hours=0, minutes=0, seconds=0.0)
    >>> TimeUnits(hours=7*24-1)
    TimeUnits(years=0, months=0, weeks=0, days=6, hours=23, minutes=0, seconds=0.0)
    >>> TimeUnits(minutes=2.4, seconds=59)
    TimeUnits(years=0, months=0, weeks=0, days=0, hours=0, minutes=3, seconds=23.0)
    >>> time_units = TimeUnits(hours=1.5, seconds=120.5)
    >>> time_units
    TimeUnits(years=0, months=0, weeks=0, days=0, hours=1, minutes=32, seconds=0.49999999999954525)
    >>> time_units.to_words()
    '1 hora y 32 minutos'
    >>> time_units.to_words(integer_seconds=False)
    '1 hora, 32 minutos y 0.49999999999954525 segundos'
    >>> time_units = TimeUnits(years=1/12, days=1, minutes=0.1)
    >>> time_units
    TimeUnits(years=0, months=1, weeks=0, days=1, hours=0, minutes=0, seconds=6.0)
    >>> time_units.to_words()
    '1 mes, 1 día y 6 segundos'
    >>> time_units = TimeUnits(days=170)
    >>> time_units
    TimeUnits(years=0, months=5, weeks=2, days=3, hours=21, minutes=59, seconds=54.23999999806938)
    >>> time_units.to_words()
    '5 meses, 2 semanas, 3 días, 21 horas, 59 minutos y 54 segundos'
    >>> time_units.to_years()
    0.4657532204917389
    >>> time_units.to_months()
    5.5890386459008665
    >>> time_units.to_weeks()
    24.285714285714285
    >>> time_units.to_days()
    170.0
    >>> time_units.to_hours()
    4080.0
    >>> time_units.to_minutes()
    244800.0
    >>> time_units.to_seconds()
    14688000.0
    """

    years: int = 0
    months: int = 0
    weeks: int = 0
    days: int = 0
    hours: int = 0
    minutes: int = 0
    seconds: float = 0

    def __init__(self, years: float = 0, months: float = 0, weeks: float = 0, days: float = 0, hours: float = 0, minutes: float = 0, seconds: float = 0):
        weeks += days / 7 + hours / 168 + minutes / 10080 + seconds / 604800

        years, remainder = divmod(years, 1)
        months += remainder * 12

        months, remainder = divmod(months, 1)
        weeks += remainder * constants.WEEKS_IN_A_MONTH

        quotient, weeks = divmod(weeks, constants.WEEKS_IN_A_MONTH)
        months += quotient

        quotient, months = divmod(months, 12)
        years += quotient

        weeks, remainder = divmod(weeks, 1)
        days = _round_if_close_to_unit(remainder * 7)

        days, remainder = divmod(days, 1)
        hours = _round_if_close_to_unit(remainder * 24)

        hours, remainder = divmod(hours, 1)
        minutes = _round_if_close_to_unit(remainder * 60)

        minutes, remainder = divmod(minutes, 1)
        seconds = _round_if_close_to_unit(remainder * 60)

        self.years = int(years)
        self.months = int(months)
        self.weeks = int(weeks)
        self.days = int(days)
        self.hours = int(hours)
        self.minutes = int(minutes)
        self.seconds = seconds

    def to_years(self) -> float:
        return self.to_months() / 12

    def to_months(self) -> float:
        return self.to_weeks() / constants.WEEKS_IN_A_MONTH

    def to_weeks(self) -> float:
        return self.to_days() / 7

    def to_days(self) -> float:
        return self.to_hours() / 24

    def to_hours(self) -> float:
        return self.to_minutes() / 60

    def to_minutes(self) -> float:
        return self.to_seconds() / 60

    def to_seconds(self) -> float:
        months = 12 * self.years
        weeks = constants.WEEKS_IN_A_MONTH * (self.months + months)
        days = 7 * (self.weeks + weeks)
        hours = 24 * (self.days + days)
        minutes = 60 * (self.hours + hours)
        seconds = 60 * (self.minutes + minutes)

        return self.seconds + seconds

    total_seconds = to_seconds

    def to_words(self, language='es', integer_seconds=True) -> str:
        if language == 'es':
            translation = {
                'years': {'singular': 'año', 'plural': 'años'},
                'months': {'singular': 'mes', 'plural': 'meses'},
                'weeks': {'singular': 'semana', 'plural': 'semanas'},
                'days': {'singular': 'día', 'plural': 'días'},
                'hours': {'singular': 'hora', 'plural': 'horas'},
                'minutes': {'singular': 'minuto', 'plural': 'minutos'},
                'seconds': {'singular': 'segundo', 'plural': 'segundos'}
            }
            separator = ', '
            last_separator = ' y '
        elif language == 'en':
            translation = {
                'years': {'singular': 'year', 'plural': 'years'},
                'months': {'singular': 'month', 'plural': 'months'},
                'weeks': {'singular': 'week', 'plural': 'weeks'},
                'days': {'singular': 'day', 'plural': 'days'},
                'hours': {'singular': 'hour', 'plural': 'hours'},
                'minutes': {'singular': 'minute', 'plural': 'minutes'},
                'seconds': {'singular': 'second', 'plural': 'seconds'}
            }
            separator = ', '
            last_separator = ' and '
        else:
            raise NotImplementedError('not implemented for that language')

        self_vars = vars(self).copy()
        if integer_seconds:
            self_vars['seconds'] = int(self_vars['seconds'])
        words = (f"{v} {translation[k]['singular'] if v == 1 else translation[k]['plural']}" for k, v in self_vars.items() if v)
        return strings.join_last_separator(words, separator, last_separator)



import asyncio
import functools
import html
import random

import aiohttp
import aiohttp.client_exceptions
import browser_cookie3
import yarl

from flanautils import constants
from flanautils.exceptions import ResponseError
from flanautils.models.enums import HTTPMethod


def browser_cookies(domain: str, ignore_expired=True) -> list[dict]:
    """Obtains chrome cookies according to domain parameter."""

    cookies = []

    for cookie in browser_cookie3.chrome(domain_name=domain):
        cookie_vars = vars(cookie)
        if ignore_expired and 'expires' in cookie_vars and not cookie_vars['expires']:
            continue

        cookie_dict = {}
        for k, v in cookie_vars.items():
            if k.startswith('_'):
                cookie_dict |= v
                continue

            if k == 'secure':
                v = bool(v)
            cookie_dict[k] = v
        cookies.append(cookie_dict)

    return cookies


async def request(
    http_method: HTTPMethod,
    url: str,
    params: dict = None,
    headers: dict = None,
    data: dict = None,
    session: aiohttp.ClientSession = None,
    clean_text=True,
    return_response=False,
    attempts=5
) -> bytes | str | list | dict | aiohttp.ClientResponse:
    """
    Function that simplifies asynchronous http requests with aiohttp.

    If return_response=True it returns the response object instead of the response data (by default return_response=False).

    Retry the request if it fails up to the number of times specified by tries (by default tries=5).

    Raise exceptions.ResponseError if response.status != 200.
    """

    if not url.startswith('http'):
        url = f'https://{url}'

    if params:
        params = {str(k): str(v) for k, v in params.items()}

    session_ = session or aiohttp.ClientSession()

    try:
        if http_method is HTTPMethod.GET:
            http_method = session_.get
        elif http_method is HTTPMethod.POST:
            http_method = session_.post
        else:
            raise ValueError('Bad http method.')

        for attempt in range(attempts - 1, -1, -1):
            try:
                async with http_method(yarl.URL(url, encoded=True), params=params, headers=headers, data=data) as response:
                    if return_response:
                        return response
                    if response.status != 200:
                        raise ResponseError(f'{response.status} - {response.reason} - {await response.read()}')

                    if response.content_type == 'application/json':
                        return await response.json()
                    elif 'text' in response.content_type:
                        if clean_text:
                            return html.unescape((await response.read()).decode('unicode_escape').encode(errors='xmlcharrefreplace').decode().replace('\\', ''))
                        else:
                            return await response.text()
                    else:
                        return await response.read()
            except aiohttp.client_exceptions.ServerDisconnectedError:
                if not attempt:
                    raise
                await asyncio.sleep(1)
    finally:
        if not session:
            await session_.close()


get_request = functools.partial(request, HTTPMethod.GET)
post_request = functools.partial(request, HTTPMethod.POST)


async def resolve_real_url(url: str, headers=None) -> str:
    """Gets the final url after the redirects."""

    if headers is None:
        headers = {'User-Agent': random.choice(constants.GOOGLE_BOT_USER_AGENTS)}
    return str((await get_request(url, headers=headers, return_response=True)).url)



import contextlib
import os
import pathlib
import pkgutil
import subprocess
import sys
from collections.abc import Iterator
from contextlib import contextmanager
from typing import overload

from flanautils import strings


def find_paths_by_stem(
    stem: str,
    directory: str | pathlib.Path = '',
    lazy=False
) -> Iterator[pathlib.Path] | list[pathlib.Path]:
    """
    Returns the pathlib.Path objects of the directory that have the stem.

    If lazy=False (the default) it returns a list, if lazy=True, returns a generator.
    """

    generator_ = (path for path in pathlib.Path(directory).iterdir() if str(path).split('.', maxsplit=1)[0] == stem)
    return generator_ if lazy else list(generator_)


@overload
def next_path(path_template: str, start=1, exhaustive=True) -> str:
    pass


@overload
def next_path(path_template: pathlib.Path, start=1, exhaustive=True) -> pathlib.Path:
    pass


def next_path(path_template: str | pathlib.Path, start=1, exhaustive=True) -> str | pathlib.Path:
    """
    Finds the next free path in a sequentially named files.

    path_template has to follow the syntax of str.format().

    If exhaustive=True (the default) the function will take into account the gaps and ensure the lowest available index.
    If not, it will use a more efficient algorithm to find a free index, but it will not necessarily be the lowest
    possible.

    >>> pathlib.Path('file_1.txt').touch()
    >>> pathlib.Path('file_2.txt').touch()
    >>> pathlib.Path('file_4.txt').touch()

    >>> next_path('file_{}.txt')
    'file_3.txt'
    >>> next_path('file_{}.txt', exhaustive=False)
    'file_5.txt'

    >>> pathlib.Path('file_1.txt').unlink()
    >>> pathlib.Path('file_2.txt').unlink()
    >>> pathlib.Path('file_4.txt').unlink()
    """

    b = start

    if exhaustive:
        while pathlib.Path(path_template.format(b)).exists():
            b += 1
    else:
        while pathlib.Path(path_template.format(b)).exists():
            b *= 2
        a, b = (b // 2, b)
        while a + 1 < b:
            c = (a + b) // 2
            a, b = (c, b) if pathlib.Path(path_template.format(c)).exists() else (a, c)

    return path_template.format(b)


def resolve_path(path: str) -> pathlib.Path:
    """Resolves actual file or directory paths in distributed libraries."""

    path = path.strip('/')
    try:
        path, file = path.rsplit('/', 1)
    except ValueError:
        file = ''
    path = path.replace('/', '.')

    return pathlib.Path(pkgutil.resolve_name(path).__path__[0]) / file


def set_windows_environment_variables(variables: str | dict | pathlib.Path, search_jsons=True, set_in_system=False):
    """
    Set environment variables for Windows.

    If variables is a str or a pathlib.Path it looks for environment variables in that file in .env or json format.
    """

    match variables:
        case str() | pathlib.Path() as text:
            if pathlib.Path(text).is_file():
                with open(text) as file:
                    text = file.read()

            variables = {}
            if search_jsons:
                variables = {k: v for dict_ in strings.find_jsons(text) for k, v in dict_.items()}
            if not variables:
                variables = strings.find_environment_variables(text)
        case dict(variables):
            pass
        case _:
            raise TypeError('bad arguments')

    if set_in_system:
        for k, v in variables.items():
            subprocess.run(f'setx /m {k} {v}')
    else:
        for k, v in variables.items():
            os.environ[k] = v


@contextmanager
def suppress_low_level_stderr():
    """A context manager that redirects low level stderr to devnull."""

    with open(os.devnull, 'w') as err_null_file:
        stderr_fileno = sys.stderr.fileno()
        old_stderr_fileno = os.dup(sys.stderr.fileno())
        old_stderr = sys.stderr

        os.dup2(err_null_file.fileno(), stderr_fileno)
        sys.stderr = err_null_file

        try:
            yield
        finally:
            sys.stderr = old_stderr
            os.dup2(old_stderr_fileno, stderr_fileno)

            os.close(old_stderr_fileno)


@contextmanager
def suppress_low_level_stdout():
    """A context manager that redirects low level stdout to devnull."""

    with open(os.devnull, 'w') as out_null_file:
        stdout_fileno = sys.stdout.fileno()
        old_stdout_fileno = os.dup(sys.stdout.fileno())
        old_stdout = sys.stdout

        os.dup2(out_null_file.fileno(), stdout_fileno)
        sys.stdout = out_null_file

        try:
            yield
        finally:
            sys.stdout = old_stdout
            os.dup2(old_stdout_fileno, stdout_fileno)

            os.close(old_stdout_fileno)


@contextmanager
def suppress_stderr():
    """A context manager that redirects stderr to devnull."""

    with open(os.devnull, 'w') as null_file, contextlib.redirect_stderr(null_file):
        yield


@contextmanager
def suppress_stdout():
    """A context manager that redirects stdout to devnull."""

    with open(os.devnull, 'w') as null_file, contextlib.redirect_stdout(null_file):
        yield



from __future__ import annotations  # todo0 remove when it's by default

import itertools
import pathlib
from dataclasses import dataclass
from enum import Enum
from typing import Any, overload

from flanautils import iterables
from flanautils.models.bases import FlanaBase
from flanautils.models.enums import MediaType, Source


@dataclass(unsafe_hash=True)
class Media(FlanaBase):
    url: str = None
    bytes_: bytes = None
    type_: MediaType = None
    extension: str = None
    source: str | Source = None
    title: str = None
    author: str = None
    album: str = None
    song_info: Media = None

    @overload
    def __init__(self, url: str | pathlib.Path = None, type_: MediaType = None, extension: str = None, source: str | Source = None, title: str = None, author: str = None, album: str = None, song_info: Media = None):
        pass

    @overload
    def __init__(self, bytes_: bytes = None, type_: MediaType = None, extension: str = None, source: str | Source = None, title: str = None, author: str = None, album: str = None, song_info: Media = None):
        pass

    @overload
    def __init__(self, url: str | pathlib.Path = None, bytes_: bytes = None, type_: MediaType = None, extension: str = None, source: str | Source = None, title: str = None, author: str = None, album: str = None, song_info: Media = None):
        pass

    @overload
    def __init__(self, bytes_: bytes = None, url: str | pathlib.Path = None, type_: MediaType = None, extension: str = None, source: str | Source = None, title: str = None, author: str = None, album: str = None, song_info: Media = None):
        pass

    def __init__(self, *args, **kwargs):
        main_ars = list(itertools.takewhile(lambda arg_: isinstance(arg_, str | bytes | pathlib.Path), args))
        self.url = iterables.find(main_ars, str | pathlib.Path)
        self.bytes_ = iterables.find(reversed(main_ars), bytes)

        rest_attributes_names = ('type_', 'extension', 'source', 'title', 'author', 'album', 'song_info')
        for attributes_name, arg in zip(rest_attributes_names, itertools.dropwhile(lambda arg_: isinstance(arg_, str | bytes), args)):
            setattr(self, attributes_name, arg)

        self.url = kwargs.get('url') or self.url
        if isinstance(self.url, pathlib.Path):
            self.url = str(self.url)
        self.bytes_ = kwargs.get('bytes_') or self.bytes_
        self.type_ = kwargs.get('type_') or self.type_
        self.extension = kwargs.get('extension') or self.extension
        if not self.extension and self.type_:
            self.extension = self.type_.default_extension
        self.source = kwargs.get('source') or self.source
        self.title = kwargs.get('title') or self.title
        self.author = kwargs.get('author') or self.author
        self.album = kwargs.get('album') or self.album
        self.song_info = kwargs.get('song_info') or self.song_info

    def __bool__(self):
        return bool(self.content)

    def _json_repr(self) -> Any:
        return {k: v.name if isinstance(v, Enum) else v for k, v in super()._json_repr().items()}

    @property
    def content(self) -> str | bytes | None:
        return self.url or self.bytes_

    def is_audio(self) -> bool:
        return self.type_ is MediaType.AUDIO

    def is_gif(self) -> bool:
        return self.type_ is MediaType.GIF

    def is_image(self) -> bool:
        return self.type_ is MediaType.IMAGE

    def is_video(self) -> bool:
        return self.type_ is MediaType.VIDEO



from flanautils.constants import *
from flanautils.asyncs import *
from flanautils.data_structures import *
from flanautils.exceptions import *
from flanautils.functions import *
from flanautils.images import *
from flanautils.iterables import *
from flanautils.maths import *
from flanautils.medias import *
from flanautils.models import *
from flanautils.oss import *
from flanautils.requests import *
from flanautils.strings import *



import asyncio
import datetime
import inspect
import multiprocessing
import queue
from asyncio import Task
from collections.abc import Callable, Iterable
from typing import Any, Type


async def _do(
    func: Callable,
    *args,
    exceptions_to_capture: Type[Exception] | Iterable[Type[Exception]] = (),
    **kwargs
) -> Any:
    """Execute the function whether it is asynchronous or not."""

    try:
        return await result if inspect.isawaitable(result := func(*args, **kwargs)) else result
    except (*exceptions_to_capture,) as e:
        return e


def _process_function(
    func: Callable,
    *args,
    queue__: multiprocessing.Queue,
    **kwargs
):
    """Utility function for multiprocessing purposes."""

    if inspect.isawaitable(result := func(*args, **kwargs)):
        queue__.put(asyncio.run(result))
    else:
        queue__.put(result)


def do_every(
    seconds: int | float | datetime.timedelta,
    func: Callable,
    *args,
    times: int = None,
    do_first_now=True,
    exceptions_to_capture: Type[Exception] | Iterable[Type[Exception]] = (),
    **kwargs
) -> Task:
    """
    Corutine function that runs the function every time provided in seconds or datetime.timedelta.

    You can specify a number of times.

    If you want to run the function the first time before waiting set do_first_now=True (the default).

    Exceptions specified in exceptions_to_capture are returned if raised when the task is completed.
    """

    async def do_every_(times_: int):
        if times_ is None:
            if do_first_now:
                await _do(func, *args, exceptions_to_capture=exceptions_to_capture, **kwargs)
            while True:
                await asyncio.sleep(seconds)
                await _do(func, *args, exceptions_to_capture=exceptions_to_capture, **kwargs)
        else:
            results = []
            if do_first_now and times > 0:
                results.append(await _do(func, *args, exceptions_to_capture=exceptions_to_capture, **kwargs))
                times_ -= 1
            for _ in range(times_):
                await asyncio.sleep(seconds)
                results.append(await _do(func, *args, exceptions_to_capture=exceptions_to_capture, **kwargs))
            return results

    if isinstance(seconds, datetime.timedelta):
        seconds = seconds.total_seconds()
    if not isinstance(exceptions_to_capture, Iterable):
        exceptions_to_capture = (exceptions_to_capture,)

    return asyncio.create_task(do_every_(times))


def do_later(
    seconds: int | float | datetime.timedelta,
    func: Callable,
    *args,
    exceptions_to_capture: Type[Exception] | Iterable[Type[Exception]] = (),
    **kwargs
) -> Task:
    """
    Corutine function that executes the function after a time provided in seconds or datetime.timedelta.

    Exceptions specified in exceptions_to_capture are returned if raised.
    """

    async def do_later_():
        await asyncio.sleep(seconds)
        return await _do(func, *args, exceptions_to_capture=exceptions_to_capture, **kwargs)

    if isinstance(seconds, datetime.timedelta):
        seconds = seconds.total_seconds()
    if not isinstance(exceptions_to_capture, Iterable):
        exceptions_to_capture = (exceptions_to_capture,)

    return asyncio.create_task(do_later_())


async def poll_process(process: multiprocessing.Process, sleep_seconds=1):
    """
    Starts the process and wait until the process is done.

    Checks every sleep_seconds (1 by default) if the process has finished.
    """

    process.start()
    while process.is_alive():
        await asyncio.sleep(sleep_seconds)


async def run_process(func: Callable, *args, timeout: int | float = None, **kwargs) -> Any:
    """
    Executes the function with the provided arguments in another process in parallel, waits asynchronously for its
    completion, and returns the result.

    If the timeout seconds expire an asyncio.TimeoutError is raised.
    """

    queue__ = multiprocessing.Queue()
    await wait_for_process(
        multiprocessing.Process(
            target=_process_function,
            args=(func, *args),
            kwargs={'queue__': queue__, **kwargs}
        ),
        timeout
    )
    try:
        return queue__.get(block=False)
    except queue.Empty:
        pass


async def wait_for_process(process: multiprocessing.Process, timeout: int | float = None):
    """
    Wrapper function that starts the process and wait until the process is done.

    If the timeout seconds expire the process is terminated and an asyncio.TimeoutError is raised.
    """

    try:
        await asyncio.wait_for(poll_process(process), timeout)
    except asyncio.TimeoutError:
        process.terminate()
        raise



import os

import pymongo

from flanautils.models import MongoBase

mongo_client = None
database = None


def init_database(reload=False):
    global mongo_client
    global database

    if mongo_client and not reload:
        return

    mongo_client = pymongo.MongoClient(
        host=os.environ.get('MONGO_HOST'),
        port=int(port) if (port := os.environ.get('MONGO_PORT')) else None,
        username=os.environ.get('MONGO_USER'),
        password=os.environ.get('MONGO_PASSWORD'),
        tz_aware=True
    )

    if database_name := os.environ.get('DATABASE_NAME'):
        database = mongo_client[database_name]
        MongoBase.init_database_attributes(database)



import asyncio
import pathlib
import uuid

import flanautils


async def edit_metadata(input_file: bytes | str | pathlib.Path, metadata: dict, overwrite=True) -> bytes:
    """Edits the media file metadata."""

    if not overwrite:
        old_metadata = await get_metadata(input_file)
        metadata = {k: v for k, v in metadata.items() if k not in old_metadata}
    if not metadata:
        if isinstance(input_file, bytes):
            return input_file
        else:
            return pathlib.Path(input_file).read_bytes()

    metadata_args = []
    for k, v in metadata.items():
        metadata_args.append('-metadata')
        metadata_args.append(f'{k}={v}')

    if isinstance(input_file, bytes):
        input_file_name = str(uuid.uuid1())
        input_file_path = pathlib.Path(input_file_name)
        input_file_path.write_bytes(input_file)
    else:
        input_file_name = str(input_file)
        input_file_path = pathlib.Path(input_file)

    if not (extension := input_file_path.suffix.strip('.')):
        extension = await get_format(input_file)
        if 'mp4' in extension:
            extension = 'mp4'

    output_file_name = f'{str(uuid.uuid1())}.{extension}'

    process = await asyncio.create_subprocess_exec(
        'ffmpeg', '-i', input_file_name, '-c', 'copy', *metadata_args, output_file_name,
        stdin=asyncio.subprocess.DEVNULL,
        stdout=asyncio.subprocess.DEVNULL,
        stderr=asyncio.subprocess.DEVNULL
    )
    await process.wait()

    if isinstance(input_file, bytes):
        input_file_path.unlink(missing_ok=True)

    output_file_path = pathlib.Path(output_file_name)
    bytes_ = output_file_path.read_bytes()
    output_file_path.unlink(missing_ok=True)

    return bytes_


async def get_format(input_file: bytes | str | pathlib.Path) -> str:
    """Gets media file format."""

    if isinstance(input_file, bytes):
        input_file_name = str(uuid.uuid1())
        input_file_path = pathlib.Path(input_file_name)
        input_file_path.write_bytes(input_file)
    else:
        input_file_name = str(input_file)

    process = await asyncio.create_subprocess_exec(
        'ffprobe', '-show_format', input_file_name,
        stdin=asyncio.subprocess.DEVNULL,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE
    )
    stdout, _stderr = await process.communicate()

    if isinstance(input_file, bytes):
        # noinspection PyUnboundLocalVariable
        input_file_path.unlink(missing_ok=True)

    if not stdout:
        raise ValueError('empty ffmpeg stdout')

    return flanautils.find_environment_variables(stdout.decode())['format_name']


async def get_metadata(input_file: bytes | str | pathlib.Path) -> dict:
    """Gets the metadata dictionary of the media file."""

    if isinstance(input_file, bytes):
        input_file_name = str(uuid.uuid1())
        input_file_path = pathlib.Path(input_file_name)
        input_file_path.write_bytes(input_file)
    else:
        input_file_name = str(input_file)

    process = await asyncio.create_subprocess_exec(
        'ffmpeg', '-i', input_file_name, '-f', 'ffmetadata', 'pipe:',
        stdin=asyncio.subprocess.DEVNULL,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE
    )
    stdout, _stderr = await process.communicate()

    if isinstance(input_file, bytes):
        # noinspection PyUnboundLocalVariable
        input_file_path.unlink(missing_ok=True)

    if not stdout:
        raise ValueError('empty ffmpeg stdout')

    metadata = flanautils.find_environment_variables(stdout.decode())

    return {k.lower(): v for k, v in metadata.items()}


# noinspection PyUnboundLocalVariable
async def merge(
    input_file_1: bytes | str | pathlib.Path,
    input_file_2: bytes | str | pathlib.Path = None,
    format_='mp4'
) -> bytes | None:
    """
    Merges the input files into one.

    If output_file is not provided returns a default_format file bytes.
    """

    if isinstance(input_file_1, bytes):
        input_file_name_1 = str(uuid.uuid1())
        input_file_path_1 = pathlib.Path(input_file_name_1)
        input_file_path_1.write_bytes(input_file_1)
    else:
        input_file_name_1 = str(input_file_1)

    if input_file_2:
        if isinstance(input_file_2, bytes):
            input_file_name_2 = str(uuid.uuid1())
            input_file_path_2 = pathlib.Path(input_file_name_2)
            input_file_path_2.write_bytes(input_file_2)
        else:
            input_file_name_2 = str(input_file_2)
        input_2_args = ('-i', input_file_name_2)
    else:
        input_2_args = ()

    output_file_name = f'{str(uuid.uuid1())}.{format_}'

    process = await asyncio.create_subprocess_exec(
        'ffmpeg', '-i', input_file_name_1, *input_2_args, '-y', '-c', 'copy', output_file_name,
        stdin=asyncio.subprocess.DEVNULL,
        stdout=asyncio.subprocess.DEVNULL,
        stderr=asyncio.subprocess.DEVNULL
    )
    await process.wait()

    if isinstance(input_file_1, bytes):
        input_file_path_1.unlink(missing_ok=True)

    if isinstance(input_file_2, bytes):
        input_file_path_2.unlink(missing_ok=True)

    output_file_path = pathlib.Path(output_file_name)
    bytes_ = output_file_path.read_bytes()
    output_file_path.unlink(missing_ok=True)

    return bytes_


async def to_gif(input_file: bytes | str | pathlib.Path) -> bytes:
    """Convert video to gif."""

    if await get_format(input_file) == 'gif':
        if isinstance(input_file, bytes):
            return input_file
        else:
            return pathlib.Path(input_file).read_bytes()

    if isinstance(input_file, bytes):
        input_file_name = str(uuid.uuid1())
        input_file_path = pathlib.Path(input_file_name)
        input_file_path.write_bytes(input_file)
    else:
        input_file_name = str(input_file)

    process = await asyncio.create_subprocess_exec(
        'ffmpeg', '-i', input_file_name, '-vf', 'fps=30,split[s0][s1];[s0]palettegen[p];[s1][p]paletteuse', '-loop', '0', '-f', 'gif', 'pipe:',
        stdin=asyncio.subprocess.DEVNULL,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE
    )
    stdout, _stderr = await process.communicate()

    if isinstance(input_file, bytes):
        # noinspection PyUnboundLocalVariable
        input_file_path.unlink(missing_ok=True)

    if not stdout:
        raise ValueError('empty ffmpeg stdout')

    return stdout


async def to_mp3(input_file: bytes | str | pathlib.Path, bitrate=192, sample_rate=44100, channels=2) -> bytes:
    """Extract and return audio in mp3 format from the media file."""

    if await get_format(input_file) == 'mp3':
        if isinstance(input_file, bytes):
            return input_file
        else:
            return pathlib.Path(input_file).read_bytes()

    if isinstance(input_file, bytes):
        input_file_name = str(uuid.uuid1())
        input_file_path = pathlib.Path(input_file_name)
        input_file_path.write_bytes(input_file)
    else:
        input_file_name = str(input_file)

    process = await asyncio.create_subprocess_exec(
        'ffmpeg', '-i', input_file_name, '-b:a', f'{bitrate}k', '-ar', str(sample_rate), '-ac', str(channels), '-f', 'mp3', 'pipe:',
        stdin=asyncio.subprocess.DEVNULL,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE
    )
    stdout, _stderr = await process.communicate()

    if isinstance(input_file, bytes):
        # noinspection PyUnboundLocalVariable
        input_file_path.unlink(missing_ok=True)

    if not stdout:
        raise ValueError('empty ffmpeg stdout')

    return stdout



from __future__ import annotations  # todo0 remove when it's by default

from typing import Any, Iterable, Mapping


class BiDict(dict):
    """
    Dictionary that also stores an inverted dictionary to access efficiently through hashes in both directions
    (key -> value, value -> key).

    Reimplements all dictionary methods.
    """

    def __init__(self, dict_: dict = None):
        dict_ = {} if dict_ is None else dict_
        super().__init__(dict_)
        self.inverted = {v: k for k, v in dict_.items()}

    def __contains__(self, item):
        if exists := super().__contains__(item):
            return exists
        else:
            return item in self.inverted

    def __delitem__(self, item):
        value = self[item]
        try:
            super().__delitem__(item)
        except (KeyError, TypeError):
            del self.inverted[value]

    def __getitem__(self, item):
        try:
            return super().__getitem__(item)
        except (KeyError, TypeError):
            return self.inverted[item]

    def __or__(self, other):
        return BiDict(super().__or__(other))

    def __ror__(self, other):
        return BiDict(dict.__or__(other, self))

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f"{{{', '.join(f'{repr(k)}::{repr(v)}' for k, v in self.items())}}}"

    def clear(self):
        super().clear()
        self.inverted.clear()

    def copy(self) -> BiDict:
        return BiDict(self)

    @classmethod
    def fromkeys(cls, iterable: Iterable, value: Any = None) -> BiDict:
        return cls(dict.fromkeys(iterable, value))

    def get(self, key: Any, default: Any = None) -> Any:
        if (value := super().get(key, default)) == default:
            return self.inverted.get(key, default)
        else:
            return value

    def pop(self, key: Any) -> Any:
        try:
            return super().pop(key)
        except KeyError:
            return self.inverted.pop(key)

    def setdefault(self, key: Any, default: Any = None) -> Any:
        if key in self:
            return self[key]
        else:
            self[key] = default
            self.inverted[default] = key
            return default

    def union(self, mapping: Mapping | Iterable, **kwargs):
        new_bi_dict = BiDict(self)
        new_bi_dict.update(mapping, **kwargs)
        return new_bi_dict

    def update(self, mapping: Mapping | Iterable, **kwargs):
        if isinstance(mapping, Mapping):
            items = list(mapping.items())
        elif isinstance(mapping, Iterable):
            items = [(element[0], element[1]) for element in mapping]
        else:
            items = []
        items.extend(list(kwargs.items()))

        for k, v in items:
            self[k] = v
            self.inverted[v] = k

    union_update = update
